! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

 module slib_basic_math

  implicit none
  
  private
  
  public :: hermite_poly, l_polynomial, hermite_integrate2D


contains 



  subroutine hermite_integrate2D(n, beta, x0, x1, y0, y1, xcenter, ycenter, Ix, Iy)

    implicit none
    
    integer n
    DOUBLE PRECISION beta, x0, x1, y0, y1, xcenter, ycenter
    DOUBLE PRECISION, DIMENSION(0:n) :: Ix,Iy  

    call hermite_intergrate ( n, beta, x0, x1, xcenter, Ix )
    call hermite_intergrate ( n, beta, y0, y1, ycenter, Iy )

    !call hermite_direct_intergrate(n, beta, x0, x1, xcenter, Ix)
    !call hermite_direct_intergrate(n, beta, y0, y1, ycenter, Iy)

end subroutine hermite_integrate2D
 




!****************************************************************************
  subroutine hermite_direct_intergrate(n_hermite, beta_shape, x0, x1, xcenter, ix)
!****************************************************************************

    use slib_constants
    implicit none
    
    integer n_hermite
    double precision beta_shape
    double precision x0
    double precision x1
    double precision xcenter
    double precision ix(0:n_hermite)
    
    integer i, i_hermite, NSAMPLE
    double precision herm
    double precision c_hermite(0:n_hermite)
    double precision x, dx, normx, xprime, sum

  
    NSAMPLE = 100

    dx = (x1-x0)/dble(NSAMPLE)

    ix=0.0

    do i=0,NSAMPLE

       x = x0 + i * dx

       xprime = (x-xcenter) / beta_shape
       
       call hermite_poly ( n_hermite, xprime, c_hermite)
       
       
       do i_hermite=0, n_hermite		
          

          normx=1. / sqrt( 2**i_hermite * sqrt(myPI) * GAMMA(real(i_hermite+1)) ) 
          normx=normx*1./sqrt(beta_shape)*exp(-xprime**2/2) * c_hermite(i_hermite)
          ix(i_hermite) = ix(i_hermite) + normx * dx

       end do
    end do

end subroutine hermite_direct_intergrate



  
  subroutine hermite_intergrate ( n, beta, x0, x1, xcenter, ix )

    !****************************************************************************80
    !
    !  Purpose:
    !
    !    Integration of hermite coefficients between x0 and x1
    !
    !
    !    I0(x) = sqrt( beta * sqrt(pi) / 2 ) * [erf(x)]_{x0}^{x1}
    !    I1(x) = -sqrt(2 * beta ) * [Hermite_0(x)]_{x0}^{x1}
    !    IN(x) = -beta * sqrt(2 / N ) * [Hermite_N-1(x)]_{x0}^{x1} + sqrt[(N-1)/N)] * IN-2(X)

    use slib_constants
    
    implicit none
    
    integer n
    DOUBLE PRECISION beta, x0, x1, xcenter
    DOUBLE PRECISION, DIMENSION(0:n) :: ix
    
    integer i
    DOUBLE PRECISION  herm, herm_x0, herm_x1, xprime0, xprime1, xprime0p, xprime1p
    double precision, DIMENSION(0:n) ::  cx_x0, cx_x1

    xprime0 = (x0 - xcenter ) / beta
    xprime1 = (x1 - xcenter ) / beta
  
    call hermite_poly ( n, xprime0, cx_x0 )
    call hermite_poly ( n, xprime1, cx_x1 )

    if ( n < 0 )then
       return
    end if

  ! these are phi_{n=0)(x1), phi_{n=0}(x0)
    herm = 1.0 / sqrt( sqrt(myPI) * beta )
    herm_x1 = herm * exp(-xprime1**2/2.0) * cx_x1(0)  
    herm_x0 = herm * exp(-xprime0**2/2.0) * cx_x0(0)  

    xprime0p = xprime0 / sqrt(2.0)
    xprime1p = xprime1 / sqrt(2.0)
    
    ix(0) = sqrt(beta * sqrt(myPI) / 2.0 ) * (erf(xprime1p) - erf(xprime0p))
    ix(1) = -beta * sqrt(2.0) * (herm_x1 - herm_x0)

    !std::cout<<"0: "<<ix[0]<<" x0=["<<x0<<"-"<<xcenter<<"]/"<<beta<<std::endl
    !std::cout<<"1: "<<ix[1]<<" x1=["<<x1<<"-"<<xcenter<<"]/"<<beta<<std::endl
  
    do i = 2, n
       !there are phi_{n-1}(x0), phi_{n-1}(X1)
       herm =1. / sqrt( 2**(i-1) * sqrt(myPI) * GAMMA(real(i)) * beta)
       herm_x0 =herm * exp(-xprime0**2/2)*cx_x0(i-1)     
       herm_x1 =herm * exp(-xprime1**2/2)*cx_x1(i-1) 
    
       ix(i) = -beta * sqrt(2.0/i) * (herm_x1 - herm_x0 ) + sqrt( (i - 1.0) / i ) * ix(i-2) 
    enddo

  end subroutine hermite_intergrate

  

  subroutine hermite_poly ( n, x, cx )

!*****************************************************************************80
!
!! HERMITE_POLY evaluates the physicist's Hermite polynomials at X.
!
!  Differential equation:
!
!    Y'' - 2 X Y' + 2 N Y = 0
!
!  First terms:
!
!      1
!      2 X
!      4 X^2     -  2
!      8 X^3     - 12 X
!     16 X^4     - 48 X^2     + 12
!     32 X^5    - 160 X^3    + 120 X
!     64 X^6    - 480 X^4    + 720 X^2    - 120
!    128 X^7   - 1344 X^5   + 3360 X^3   - 1680 X
!    256 X^8   - 3584 X^6  + 13440 X^4  - 13440 X^2   + 1680
!    512 X^9   - 9216 X^7  + 48384 X^5  - 80640 X^3  + 30240 X
!   1024 X^10 - 23040 X^8 + 161280 X^6 - 403200 X^4 + 302400 X^2 - 30240
!
!  Recursion:
!
!    H(0,X) = 1,
!    H(1,X) = 2*X,
!    H(N,X) = 2*X * H(N-1,X) - 2*(N-1) * H(N-2,X)
!
!  Norm:
!
!    Integral ( -oo < X < oo ) exp ( - X^2 ) * H(N,X)^2 dX
!    = sqrt ( myPI ) * 2^N * N!
!
!    H(N,X) = (-1)^N * exp ( X^2 ) * dn/dXn ( exp(-X^2 ) )
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license. 
!
!  Modified:
!
!    01 October 2002
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Milton Abramowitz, Irene Stegun,
!    Handbook of Mathematical Functions,
!    National Bureau of Standards, 1964,
!    ISBN: 0-486-61272-4,
!    LC: QA47.A34.
!
!    Larry Andrews,
!    Special Functions of Mathematics for Engineers,
!    Second Edition, 
!    Oxford University Press, 1998.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the highest order polynomial to compute.
!    Note that polynomials 0 through N will be computed.
!
!    Input, real ( kind = 8 ) X, the point at which the polynomials are 
!    to be evaluated.
!
!    Output, real ( kind = 8 ) CX(0:N), the values of the first N+1 Hermite
!    polynomials at the point X.
!
  implicit none

  integer ( kind = 4 ) n

  real ( kind = 8 ) cx(0:n)
  integer ( kind = 4 ) i
  real ( kind = 8 ) x

  if ( n < 0 ) then
    return
  end if

  cx(0) = 1.0D+00

  if ( n == 0 ) then
    return
  end if

  cx(1) = 2.0D+00 * x
   !write(*,*) "1",cx(1),x

  do i = 2, n
    cx(i) = 2.0D+00 * x * cx(i-1) - 2.0D+00 * real ( i - 1, kind = 8 ) * cx(i-2)
    !write(*,*) i,cx(i),x
  end do
  
  

  return
end subroutine hermite_poly






subroutine l_polynomial ( m, n, x, v )

!*****************************************************************************80
!
!! L_POLYNOMIAL evaluates the Laguerre polynomial L(n,x).
!
!  First terms:
!
!      1
!     -X     +  1
!   (  X^2 -  4 X      +  2 ) / 2
!   ( -X^3 +  9 X^2 -  18 X    +    6 ) / 6
!   (  X^4 - 16 X^3 +  72 X^2 -   96 X +      24 ) / 24
!   ( -X^5 + 25 X^4 - 200 X^3 +  600 X^2 -   600 X    +  120 ) / 120
!   (  X^6 - 36 X^5 + 450 X^4 - 2400 X^3 +  5400 X^2 -  4320 X     + 720 ) 
!     / 720
!   ( -X^7 + 49 X^6 - 882 X^5 + 7350 X^4 - 29400 X^3 + 52920 X^2 - 35280 X 
!     + 5040 ) / 5040
!
!  Recursion:
!
!    L(0,X) = 1
!    L(1,X) = 1 - X
!    L(N,X) = (2*N-1-X)/N * L(N-1,X) - (N-1)/N * L(N-2,X)
!
!  Orthogonality:
!
!    Integral ( 0 <= X < oo ) exp ( - X ) * L(N,X) * L(M,X) dX = delta ( M, N )
!
!  Relations:
!
!    L(N,X) = (-1)^N / N! * exp ( x ) * (d/dx)^n ( exp ( - x ) * x^n )  
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license. 
!
!  Modified:
!
!    07 March 2012
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Milton Abramowitz, Irene Stegun,
!    Handbook of Mathematical Functions,
!    National Bureau of Standards, 1964,
!    ISBN: 0-486-61272-4,
!    LC: QA47.A34.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) M, the number of evaluation points.
!
!    Input, integer ( kind = 4 ) N, the highest order polynomial to compute.
!    Note that polynomials 0 through N will be computed.
!
!    Input, real ( kind = 8 ) X(M), the evaluation points.
!
!    Output, real ( kind = 8 ) V(M,0:N), the function values.
!
  implicit none

  integer ( kind = 4 ) m
  integer ( kind = 4 ) n

  integer ( kind = 4 ) j
  real ( kind = 8 ) v(m,0:n)
  real ( kind = 8 ) x(m)

  if ( n < 0 ) then
    return
  end if

  v(1:m,0) = 1.0D+00

  if ( n == 0 ) then
    return
  end if

  v(1:m,1) = 1.0D+00 - x(1:m)
 
  do j = 2, n

    v(1:m,j) = ( ( real ( 2 * j - 1, kind = 8 ) - x(1:m) ) * v(1:m,j-1)   &
                 - real (     j - 1, kind = 8 )            * v(1:m,j-2) ) &
                 / real (     j,     kind = 8 )

  end do

  return
end subroutine l_polynomial



end module slib_basic_math
