! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module slib_data_src
  
  use slib_constants
  
  implicit none

  private

  !public :: slib_init_source_psf, 
  public :: slib_deallocate_src, slib_read_src, slib_init_source_with_num, slib_read_mllist, slib_ml2src


contains



  ! this is the destructor
  subroutine slib_deallocate_src(mySRC)

    TYPE(SRC) :: mySRC
 
    IF (ALLOCATED (mySRC%RA)) DEALLOCATE (mySRC%RA)
    IF (ALLOCATED (mySRC%DEC)) DEALLOCATE (mySRC%DEC)
    IF (ALLOCATED (mySRC%RADIUS)) DEALLOCATE (mySRC%RADIUS)
 
    IF (ALLOCATED (mySRC%XD)) DEALLOCATE (mySRC%XD)
    IF (ALLOCATED (mySRC%YD)) DEALLOCATE (mySRC%YD)
    IF (ALLOCATED (mySRC%ZD)) DEALLOCATE (mySRC%ZD)

    IF (ALLOCATED (mySRC%C1)) DEALLOCATE(mySRC%C1)
    IF (ALLOCATED (mySRC%C2)) DEALLOCATE(mySRC%C2)

    IF (ALLOCATED (mySRC%XIMA)) DEALLOCATE(mySRC%XIMA)
    IF (ALLOCATED (mySRC%YIMA)) DEALLOCATE(mySRC%YIMA)


  end subroutine slib_deallocate_src


  !this is the constructor
  subroutine slib_init_source_with_num(mySRC, NUMSRC)



    implicit none

    TYPE(SRC) :: mySRC
    INTEGER NUMSRC
    
    mySRC%TOTAL = NUMSRC

    call slib_deallocate_src(mySRC)
    allocate  ( mySRC%RA(NUMSRC), mySRC%DEC(NUMSRC), mySRC%RADIUS(NUMSRC) )
    allocate  ( mySRC%XD(NUMSRC), mySRC%YD(NUMSRC), mySRC%ZD(NUMSRC) )
    allocate  ( mySRC%XIMA(NUMSRC), mySRC%YIMA(NUMSRC))

    !call slib_init_source_psf(mySRC, myMPSF)


  end subroutine slib_init_source_with_num



  subroutine slib_read_src(file, mySRC)


    implicit none

    CHARACTER(len=*) :: file

    TYPE(SRC) :: mySRC

    integer :: unit, status, rwmode, blocksize, nrows_ev


    status=0
    rwmode=0 ! open as readonly 

    call ftgiou ( unit, status )

    !write(*,*)  file

    call ftopen( unit, file, rwmode, blocksize, status)  

    !call ftmnhd(unit, 2 , 'EVENTS', 0, status)
    !call ftgnrw(unit, nrows_ev, status)  

    call slib_read_source_extension( unit,  mySRC)


  end subroutine slib_read_src




  subroutine slib_read_source_extension( unit,  mySRC)
    !this reads the event extensionm of an event file. 

    use slib_basic_calcs 
    
    implicit none


    integer unit

    TYPE(SRC) :: mySRC

    integer nrows, irow, status
    
    logical :: casesn=.false., anyf

    integer :: colra, coldec, colradius

    double precision dnullval, dvalues

    double precision ra, dec, xd, yd, zd

    real  rnullval, rvalues

    integer inullval, ivalues

    status=0
    dnullval=0.d0 
    inullval=0 

    call ftmnhd(unit, 2 , 'SOURCES', 0, status)
    call ftgnrw(unit, nrows, status)  
    
    mySRC%TOTAL = nrows

    !write(*,*) "TOTAL SOURCES", nrows


    call ftgcno(unit,casesn,'RA_DEG',colra,status)   
    call ftgcno(unit,casesn,'DEC_DEG',coldec,status)  
    call ftgcno(unit,casesn,'RAD_S',colradius,status)  

    !write(*,*) "COLS", colra, coldec, colradius
    call slib_init_source_with_num(mySRC, nrows)


    do irow=1, nrows

       call FTGCVD(unit,colra,irow,1,1,dnullval,dvalues,anyf,status)
       mySRC%RA(irow)= dvalues

       !write(*,*) irow, dvalues, mySRC%RA(irow)
       
       call FTGCVD(unit,coldec,irow,1,1,dnullval,dvalues,anyf,status)
       mySRC%DEC(irow)= dvalues

       !write(*,*) irow, dvalues, mySRC%DEC(irow)

       call FTGCVE(unit,colradius,irow,1,1,rnullval,rvalues,anyf,status)
       mySRC%RADIUS(irow)= rvalues

       ra= mySRC%RA(irow) / myRAD2DEG
       dec= mySRC%DEC(irow) / myRAD2DEG

       call  slib_src_pos2xyz ( ra, dec, xd, yd, zd)

       mySRC%XD(irow)= xd
       mySRC%YD(irow)= yd
       mySRC%ZD(irow)= zd

       !write(*,*) irow, rvalues,mySRC%RA(irow), mySRC%DEC(irow), mySRC%RADIUS(irow)


    end do

  end subroutine slib_read_source_extension
  

  !----------------------------------------------------------------------------
  subroutine slib_read_mllist(filename, mllist, ierr)
  !----------------------------------------------------------------------------
  ! reads mllist and passes it to 2d boxlist type to be consistent 
  ! with various ermldetect subroutine calls
  !----------------------------------------------------------------------------
    use detect_io

    implicit none
    type(mllist_type), intent(in) :: mllist
    character(len=*), intent(in) :: filename
    integer(kind=int32), intent(out) :: ierr

    call fget_mllist(filename,mllist,ierr)

  end subroutine slib_read_mllist


!----------------------------------------------------------------------------
  subroutine slib_ml2src(mllist, mySRC, ierr) 
!----------------------------------------------------------------------------
! convert an input mllist into a boxlist to be consistent with ermldet  
          

    use detect_io
    use slib_basic_calcs
    !use ermldet_mod

    implicit none

    type(mllist_type), intent(in) :: mllist
    integer(kind=int32), intent(out) :: ierr
    TYPE(SRC) :: mySRC

    logical(kind=4) :: has_instsum,has_bandsum   
    integer(kind=int32)  :: i,j,k,nrows,ord=1,nbands,NUMSRC
    character(len=*), parameter  :: rtname='SLIB_ML2SRC'
    type(boxlist_type) :: tmplist
    integer(kind=int32),dimension(:),allocatable    :: isort
    character(len=80)                               :: errstr
    integer(kind=int32)                             :: sla_err

    double precision ra, dec, xd, yd, zd

   ! check source list for summary rows:
    has_bandsum = .false.
    has_instsum = .false.
    nbands = maxval(mllist%data%id_band) 
    if  (minval(mllist%data%id_band) /=  maxval(mllist%data%id_band)) then
      if (minval(mllist%data%id_band) == 0) then 
        has_bandsum = .true.
      else
        ierr=3
        errstr = 'Unsupported format of input source list'
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    if  (minval(mllist%data%id_inst) /=  maxval(mllist%data%id_inst)) then
      if (minval(mllist%data%id_inst) == 0) then 
        has_instsum = .true.
      else
        ierr=3
        errstr = 'Unsupported format of input source list'
        call WRERR(rtname,ierr,errstr)
      endif
    endif

    call slib_deallocate_src(mySRC)
    nrows=mllist%shapehdr%naxis2

    
    !write(*,*) rtname, nrows, mllist%shapehdr%naxis2

    if (has_instsum) then
      j=0
      do i=1,nrows
        if (mllist%data(i)%id_inst == 0.and.mllist%data(i)%like.gt.1e2) then
          j=j+1
         endif
      enddo
      NUMSRC=j
    else if (has_bandsum) then 
      j=0
      do i=1,nrows
        if (mllist%data(i)%id_band == 0.and.mllist%data(i)%like.gt.1e2) then
          j=j+1
         endif
      enddo
      NUMSRC=j
   endif
   
   call slib_init_source_with_num(mySRC, NUMSRC)
   
   if (has_instsum) then
      j=0
      do i=1,nrows
         if (mllist%data(i)%id_inst == 0.and.mllist%data(i)%like.gt.1e2) then
            j=j+1
            mySRC%RA(j)=dble(mllist%data(i)%ra)
            mySRC%DEC(j)=dble(mllist%data(i)%dec)
            mySRC%RADIUS(j)=200.0
            mySRC%XIMA(j)=dble(mllist%data(i)%x_ima)
            mySRC%YIMA(j)=dble(mllist%data(i)%y_ima)

            ra= mySRC%RA(j) / myRAD2DEG
            dec= mySRC%DEC(j) / myRAD2DEG
            call  slib_src_pos2xyz ( ra, dec, xd, yd, zd)
            mySRC%XD(j)= xd
            mySRC%YD(j)= yd
            mySRC%ZD(j)= zd

         endif
      enddo
   else if (has_bandsum) then 
      call slib_init_source_with_num(mySRC, NUMSRC)
      j=0
      do i=1,nrows
         if (mllist%data(i)%id_band == 0.and.mllist%data(i)%like.gt.1e2) then
            j=j+1
            mySRC%RA(j)=dble(mllist%data(i)%ra)
            mySRC%DEC(j)=dble(mllist%data(i)%dec)
            mySRC%RADIUS(j)=200.0

            mySRC%XIMA(j)=dble(mllist%data(i)%x_ima)
            mySRC%YIMA(j)=dble(mllist%data(i)%y_ima)

            ra= mySRC%RA(j) / myRAD2DEG
            dec= mySRC%DEC(j) / myRAD2DEG
            call  slib_src_pos2xyz ( ra, dec, xd, yd, zd)
            mySRC%XD(j)= xd
            mySRC%YD(j)= yd
            mySRC%ZD(j)= zd

         endif
      enddo
   endif
 

   !write(*,*) NUMSRC
   !write(*,*) mySRC%RA
   !write(*,*) mySRC%DEC
   !write(*,*) mySRC%XD
   !write(*,*) mySRC%YD
   
  end subroutine slib_ml2src



  
end  module slib_data_src
